ENUM_TIMEFRAMES TFList[8] = {PERIOD_M1, PERIOD_M5, PERIOD_M15, PERIOD_M30, PERIOD_H1, PERIOD_H4, PERIOD_D1, PERIOD_W1};
enum ENUM_TIMEFRAME_INDEX {
  TF_M1  = 0,
  TF_M5  = 1,
  TF_M15 = 2,
  TF_M30 = 3,
  TF_H1  = 4,
  TF_H4  = 5,
  TF_D1  = 6,
  TF_W1  = 7,
};

//+==================================================================+
// CBasicIndicator                                                   |
//+==================================================================+
class CBasicIndicator {
protected:
  string baseSymbol;
  ENUM_TIMEFRAMES baseTF;
  
public:
  int rates;
  int prev;
  int limit;
  int barDiff;
  int barWeight;            

  bool newCandle;
  bool isTesting;
  bool start;
  int minInit;

  CBasicIndicator(void) : rates(0), prev(0), limit(0), barDiff(0), start(true), baseSymbol(_Symbol),
                                         baseTF(_Period), isTesting((bool) MQLInfoInteger(MQL_TESTER)) {}
  ~CBasicIndicator(void) {};

  bool Init(string symbol_=NULL, ENUM_TIMEFRAMES period_=PERIOD_CURRENT) {
    baseSymbol = symbol_ == NULL ? ChartSymbol() : symbol_;
    barDiff = CalcBarDiff(period_, _Period);
    baseTF  = period_;

    barWeight = barDiff == 0 ? 1 : barDiff;

    return true;
  }

  int             GetAsSeriesIndex(int bar) {return rates - bar - 1;}
  string          Symbol(void) const {return(baseSymbol);}
  ENUM_TIMEFRAMES Period(void) const {return(baseTF);}
  void            CheckNewCandle(string symbol_= NULL, ENUM_TIMEFRAMES tf_=0) {newCandle = HasNewCandle(symbol_, tf_);}
  int             BufferSize() {return rates;};

  void            SetRates(int rateNums) {
    rates     = rateNums;  
    SetLimit();
    prev      = rates;
  }
  void            SetPreviousCal(int prevNums) {prev = prevNums;}
  void            SetLimit() {
    limit = (prev <= 0 || prev > rates) ? rates - minInit : rates - prev;
    if(limit == 0) limit += 1;
  }
  void            SetMinInit(int minInit_) {minInit = minInit_;}
  
  int             GetRates() {return rates;}
  int             GetPreviousCal() {return prev;}
  int             GetLimit() {return limit;}
  int             GetCurrentBarDiff(int bar=0) {
    int currTime = (int) GetFullTimeHour(iTime(baseSymbol, 0, bar)) / 60;
    int min      = GetMinuteByPeriod(baseTF);
    return currTime % min;
  }
};

int CalcBarDiff(ENUM_TIMEFRAMES tf1_, ENUM_TIMEFRAMES tf2_=0) {
  if(tf1_ == tf2_) return 0;
  int minute1 = GetMinuteByPeriod(tf1_);
  int minute2 = GetMinuteByPeriod(tf2_);
  int tfMax   = MathMax(minute1, minute2);
  int tfMin   = MathMin(minute1, minute2);
  return int(tfMax / tfMin);
}

int GetMinuteByPeriod(ENUM_TIMEFRAMES tf) {
  switch(tf) {
    case PERIOD_M1:      return 1;
    case PERIOD_M2:      return 2;
    case PERIOD_M3:      return 3;
    case PERIOD_M4:      return 4;
    case PERIOD_M5:      return 5;
    case PERIOD_M6:      return 6;
    case PERIOD_M10:     return 10;
    case PERIOD_M12:     return 12;
    case PERIOD_M15:     return 15;
    case PERIOD_M20:     return 20;
    case PERIOD_M30:     return 30;
    case PERIOD_H1:      return 60;
    case PERIOD_H2:      return 120;
    case PERIOD_H3:      return 180;
    case PERIOD_H4:      return 240;
    case PERIOD_H6:      return 360;
    case PERIOD_H8:      return 480;
    case PERIOD_H12:     return 720;
    case PERIOD_D1:      return 1440;
    case PERIOD_W1:      return 10080;
    case PERIOD_MN1:     return 40320;
    case PERIOD_CURRENT: return GetMinuteByPeriod(_Period);
    default:             return 0;
  }
}

int CalcBarWeight(ENUM_TIMEFRAMES tf1_, ENUM_TIMEFRAMES tf2_=0) {
  int weight = CalcBarDiff(tf1_, tf2_);
  return weight == 0 ? 1 : weight;
}

bool HasNewCandle(string symbol_= NULL, ENUM_TIMEFRAMES tf_=0) {
  long lastBarTime = SeriesInfoInteger(symbol_, tf_, SERIES_LASTBAR_DATE);
  static long lastTime = lastBarTime;
  if(lastTime == lastBarTime) return false;
  lastTime = lastBarTime;
  return true;
}

datetime GetFullTimeHour(datetime currTime) {
  int hour, min;
  string timeSplit = TimeToString(currTime, TIME_MINUTES);
  string timeCurrTemp[2];
  StringSplit(timeSplit, ':', timeCurrTemp);

  hour = int(timeCurrTemp[0]);
  min  = int(timeCurrTemp[1]);

  return hour * 3600 + min * 60;
}

//+==================================================================+
// CMACD                                                             |
//+==================================================================+
class CMACD : public CBasicIndicator{
protected:
  int handleMACD;
  int fastMAPeriod;
  int slowMAPeriod;
  int appliedPrice;
  double bufferMACD[];

public:
//--- Init
  CMACD(void) : fastMAPeriod(0), slowMAPeriod(0), appliedPrice(-1) {}

  ~CMACD(void) {CMACD::DeInit();}

  bool Init(string symbol_=NULL, ENUM_TIMEFRAMES period_=PERIOD_CURRENT, int fast_ema_period=12, int slow_ema_period=26, int applied=PRICE_CLOSE)  {
    if(!CBasicIndicator::Init(symbol_, period_)) return false;

    fastMAPeriod = fast_ema_period;
    slowMAPeriod = slow_ema_period;
    appliedPrice = applied;

    minInit      = MathMax(fast_ema_period * barWeight, slow_ema_period * barWeight); 
    handleMACD   = iMACD(baseSymbol, baseTF, fast_ema_period, slow_ema_period, 1, applied);

    return(handleMACD != INVALID_HANDLE);
  }

  void DeInit() {
    IndicatorRelease(handleMACD);
    handleMACD = INVALID_HANDLE;
    ArrayFree(bufferMACD);
  }

//--- Handling
  bool CopyData(int bar) {
    int pos = GetAsSeriesIndex(bar);
    
    datetime time = iTime(baseSymbol, 0, bar);
    double val    = CopyDataHandle(handleMACD, time, 0);
    if(val == 0 || pos < 0) return false;
    
    bufferMACD[GetAsSeriesIndex(bar)] = val;

    return true;
  }
  bool CopyData(datetime time) {
    int bar = iBarShift(baseSymbol, baseTF, time);

    int pos = GetAsSeriesIndex(bar);
    
    double val = CopyDataHandle(handleMACD, time, 0);
    if(val == 0 || pos < 0) return false;
    
    bufferMACD[GetAsSeriesIndex(bar)] = val;

    return true;
  }

//--- Utilities
  int MACDFastEmaPeriod(void) const { return(fastMAPeriod); }
  int MACDSlowEmaPeriod(void) const { return(slowMAPeriod); }
  int MACDApplied(void)       const { return(appliedPrice); }
  void SetRates(int rateNums) {
    CBasicIndicator::SetRates(rateNums);
    ResizeBuffer();
  }

  void ResizeBuffer() {
    ArrayResize(bufferMACD, rates);
    if(newCandle) bufferMACD[rates-1] = 0;
  }
  double GetValue(int bar) {return bufferMACD[GetAsSeriesIndex(bar)];}
};

double CopyDataHandle(int &handle, datetime startTime, int bufferIndex=0) {
  double data[1];
  int attempt = 3;
  int copy = -1;
  while(attempt-- > 0) {
    copy = CopyBuffer(handle, bufferIndex, startTime, 1, data);
    if(copy > 0) return data[0];
    Sleep(10);
  }
  if(copy <= 0) Print(GetLastError());
  
  return 0;
}

//+==================================================================+
// CTestMACD                                                         |
//+==================================================================+
class CTestMACD : public CBasicIndicator {
protected:
  CMACD *cmacdTF[8];
  CMACD macdM1;
  CMACD macdM5;
  CMACD macdM15;
  CMACD macdM30;
  CMACD macdH1;
  CMACD macdH4;
  CMACD macdD1;
  CMACD macdW1;
  CMACD macdMN1;
  
  long lastBarTime[8];
public:
//--- Init 
  CTestMACD(void) {
    cmacdTF[TF_M1]   = &macdM1;
    cmacdTF[TF_M5]   = &macdM5;
    cmacdTF[TF_M15]  = &macdM15;
    cmacdTF[TF_M30]  = &macdM30;
    cmacdTF[TF_H1]   = &macdH1;
    cmacdTF[TF_H4]   = &macdH4;
    cmacdTF[TF_D1]   = &macdD1;
    cmacdTF[TF_W1]   = &macdW1;
    for(ENUM_TIMEFRAME_INDEX tf=TF_M1; tf <= TF_W1; ++tf) lastBarTime[tf] = SeriesInfoInteger(_Symbol, TFList[tf], SERIES_LASTBAR_DATE);
  }
  ~CTestMACD(void) {this.DeInit();}

  bool Init(string symbol_=NULL, int fast_ema_period=12, int slow_ema_period=26, int applied=PRICE_CLOSE)  {
    if(!CBasicIndicator::Init(symbol_)) return false;
    for(ENUM_TIMEFRAME_INDEX tf=TF_M1; tf <= TF_W1; ++tf) {
      if(!cmacdTF[tf].Init(symbol_, TFList[tf], fast_ema_period, slow_ema_period, applied)) return false;
      cmacdTF[tf].barDiff   = CalcBarDiff(PERIOD_M1, TFList[tf]);
      cmacdTF[tf].barWeight = CalcBarWeight(PERIOD_M1, TFList[tf]);
    }

    this.minInit = MathMax(fast_ema_period, slow_ema_period) * cmacdTF[TF_W1].barWeight;
    for(ENUM_TIMEFRAME_INDEX tf=TF_M1; tf <= TF_W1; ++tf) cmacdTF[tf].SetMinInit(this.minInit);
    return true;
  }

  void DeInit() {
    for(ENUM_TIMEFRAME_INDEX tf=TF_M1; tf <= TF_W1; ++tf) {
      delete cmacdTF[tf];
      cmacdTF[tf] = NULL;
    }
  }

  bool MacdMTFHandling() {
    for(int bar = limit-1; bar >= 0 && !IsStopped(); --bar) {
      if(bar + minInit >= rates) continue;
      datetime timeBar = iTime(baseSymbol, PERIOD_M1, bar);
      if(timeBar == NULL) continue;
      
      for(ENUM_TIMEFRAME_INDEX tf=TF_M1; tf <= TF_W1; ++tf) {
        cmacdTF[tf].CopyData(timeBar);
      }
    }

    return true;
  }

  void SetRates(int rateNums)  {
    CBasicIndicator::SetRates(rateNums);
    if(newCandle) {
      for(ENUM_TIMEFRAME_INDEX tf=TF_M1; tf <= TF_W1; ++tf) {
        cmacdTF[tf].newCandle = true;
      }
    }
    for(ENUM_TIMEFRAME_INDEX tf=TF_M1; tf <= TF_W1; ++tf) cmacdTF[tf].SetRates(rateNums);

    if(newCandle) {
      for(ENUM_TIMEFRAME_INDEX tf=TF_M5; tf <= TF_W1; ++tf) {
        if(!CheckNewCandleTF(tf, baseSymbol)) cmacdTF[tf].newCandle = false;
      }
    }
  }

  bool CheckNewCandleTF(ENUM_TIMEFRAME_INDEX tfIndex, string symbol_=NULL) {
    long lastTime = SeriesInfoInteger(symbol_, TFList[tfIndex], SERIES_LASTBAR_DATE);
    if(lastTime == lastBarTime[tfIndex]) return false;
    lastBarTime[tfIndex] = lastTime;
    return true;
  }
};

//+==================================================================+